/*
 * dnsbl_redirector.c -- dns blacklist redirectory, providing the same
 *                       functionality for squid as mod_dnsbl provides
 *                       for the apache
 *
 * (c) 2003 Dr. Andreas Mueller, Beratung und Entwicklung
 */
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <vector>
#include <map>
#include <iostream>
#include <fstream>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>

int	debug = 0;

typedef enum { IGNORE, PASS, BLOCK }	pass_t;

typedef	std::map<std::string, pass_t>		passmap_t;
typedef	std::vector<std::string>		stringvector_t;

class	Dnsbl {
	stringvector_t	suffixes;
	passmap_t	action;
	pass_t		defaultaction;
	pass_t		checkSingle(const std::string& hostname,
				const std::string& suffix);
	std::string	last;
	std::string	redirect;
public:
	Dnsbl(void) { defaultaction = PASS; }
	~Dnsbl(void) { }

	// accessors for the default action field
	pass_t	getDefaultAction(void) { return defaultaction; }
	void	setDefaultAction(pass_t d) { defaultaction = d; }

	void	setRedirect(const std::string& r) { redirect = r; }
	std::string	getRedirect(void) const { return redirect; }

	void	add(const std::string& suffix) {
		suffixes.push_back(suffix);
	}
	void	addRule(const std::string& address, pass_t a) {
		action.insert(passmap_t::value_type(address, a));
	}
	pass_t	check(const std::string& hostname);
	void	loop(void);
};

pass_t	Dnsbl::checkSingle(const std::string& hostname,
		const std::string& suffix) {
	struct hostent	*hp;
	char	**ptr;
	// perform lookup for the synthesized name
	if (NULL == (hp = gethostbyname((hostname + "." + suffix).c_str()))) {
		return IGNORE;
	}

	// turn the address into a string
	for (ptr = hp->h_addr_list; NULL != *ptr; ptr++) {
		last = std::string(inet_ntoa(*(struct in_addr *)*ptr));

		// look for the address in the map
		passmap_t::const_iterator	i = action.find(last);

		// if found, we return the action associated with it
		if (i != action.end())
			return i->second;
	}
	return IGNORE;
}

pass_t	Dnsbl::check(const std::string& hostname) {
	// clean up the last flag, or we may end up confused later
	last = std::string("");
	stringvector_t::const_iterator	v;
	for (v = suffixes.begin(); v != suffixes.end(); v++) {
		pass_t	rc = checkSingle(hostname, *v);
		if (rc != IGNORE)
			return rc;
	}
	return IGNORE;
}

void	Dnsbl::loop(void) {
	std::string	line;
	size_t		i, j;
	while (1) {
		// read a line from stdin
		char	line[20480];
		std::cin.getline(line, sizeof(line));
		std::string	hostname(line);

		// jump to the first colon, then first non / 
		i = hostname.find_first_of(":");
		i = hostname.find_first_not_of("/", i + 1);

		// if there is another /, cut the string of there
		if (std::string::npos != (j = hostname.find_first_of("/", i))) {
			hostname = hostname.substr(i, j - i);
		} else {
			hostname = hostname.substr(i);
		}

		// if there is an @ skip to it
		if (std::string::npos != (i = hostname.find_first_of("@"))) {
			hostname = hostname.substr(i + 1);
		}

		// if there is a :, cut off there, because that's only the
		// port number, and we are not interested in that
		if (std::string::npos != (i = hostname.find_first_of(":"))) {
			hostname = hostname.substr(0, i);
		}

		// check the hostname against the checker
		switch (check(hostname)) {
		case PASS:
		case IGNORE:
			std::cout << line << std::endl;
			break;
		case BLOCK:
			std::cout << redirect << last << std::endl;
			break;
		}
	}
}

int	main(int argc, char *argv[]) {
	Dnsbl	bl;
	char	*rulefilename = NULL;
	int	c;

	/* parse command line						*/
	while (EOF != (c = getopt(argc, argv, "df:s:r:")))
		switch (c) {
		case 'd':
			debug++;
			break;
		case 'f':
			rulefilename = optarg;
			break;
		case 's':
			bl.add(std::string(optarg));
			break;
		case 'r':
			bl.setRedirect(std::string(optarg));
			break;
		}

	/* read configuration file					*/
	if (rulefilename == NULL) {
		fprintf(stderr, "no configuration file specified\n");
		exit(EXIT_FAILURE);
	}
	std::ifstream	rulefile(rulefilename);
	while (!rulefile.eof()) {
		size_t	i;
		char	rule[1024];
		rulefile.getline(rule, sizeof(rule));
		std::string	r(rule);
		if (r.length() < 3)
			continue;

		// find whitespace
		if (std::string::npos ==
			(i = r.find_first_not_of("0123456789."))) {
			fprintf(stderr, "rule file syntax error: "
				"not an IP address\n");
			exit(EXIT_FAILURE);
		};
		std::string	addr = r.substr(0, i);
		if (std::string::npos == (i = r.find_first_of("pasblock"))) {
			fprintf(stderr, "rule file syntax error: "
				"pass/block argument not found\n");
			exit(EXIT_FAILURE);
		}
		std::string	action = r.substr(i);
		pass_t	act = IGNORE;
		if (action == "block") {
			act = BLOCK;
		}
		if (action == "pass") {
			act = PASS;
		}
		if (act == IGNORE) {
			fprintf(stderr, "rule file syntax: "
				"neither pass nor block\n");
			exit(EXIT_FAILURE);
		}
		bl.addRule(addr, act);
	}

	/* begin redirector loop 					*/
	bl.loop();

	/* we should never get to this point, really 			*/
	exit(EXIT_SUCCESS);
}

